==> clean_android_design-0.2.15.tar.gz: lib/src/structure/update/update_controller.dart <==
import 'package:clean_android_design/clean_android_design.dart';
import 'package:clean_android_design/src/dialog/delete_question_dialog.dart';
import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:flutter/material.dart';

class UpdateController<T extends UpdateableDatabaseItem<T>>
    extends InheritedWidget {
  final UpdateItemFrame<T> updateItemFrame;
  final Validator<T> _validator;
  final ForceUpdateKey updateNotifier;

  final IUpdateScaffoldWidgetService<T> databaseCollection;
  final ILockService? lockService;

  final Future<bool> Function(BuildContext context)? onBeforeSave;
  final void Function(BuildContext context)? onAfterSave;
  final void Function(BuildContext context, DocumentReference? itemReference)?
      onAfterInsert;
  final void Function(BuildContext context)? onAbort;
  final Future<bool> Function(BuildContext context)? onBeforeDelete;
  final List<String>? fieldsNotToUpdate;

  ReadOnlyValidator<T> get validator => _validator.readOnly();

  UpdateController({
    required Widget child,
    required this.updateItemFrame,
    required Validator<T> validator,
    required this.databaseCollection,
    required this.onBeforeSave,
    required this.onAfterSave,
    required this.onAfterInsert,
    required this.onAbort,
    required this.fieldsNotToUpdate,
    required this.onBeforeDelete,
    required this.lockService,
  })  : _validator = validator,
        updateNotifier = ForceUpdateKey(),
        super(child: child);

  @override
  bool updateShouldNotify(InheritedWidget oldWidget) => true;

  static UpdateController<T> of<T extends UpdateableDatabaseItem<T>>(
      BuildContext context) {
    final UpdateController<T>? result =
        context.dependOnInheritedWidgetOfExactType<UpdateController<T>>();
    assert(result != null, 'No UpdateController found in context');
    return result!;
  }

  static UpdateController<T>? maybeOf<T extends UpdateableDatabaseItem<T>>(
      BuildContext context) {
    return context.dependOnInheritedWidgetOfExactType<UpdateController<T>>();
  }

  Future<bool> trySave(BuildContext context) async {
    FocusNode().requestFocus();

    var result = _validator.isValid(updateItemFrame.editableItem);
    if (result.isValid) {
      return _doSave(context);
    } else {
      if (result.firstInvalidPage != null &&
          ScaffoldPageController.isIn(context)) {
        ScaffoldPageController.of(context).setPage(result.firstInvalidPage!);
        await Future.delayed(Duration(milliseconds: 100));
        result = _validator.isValid(updateItemFrame.editableItem);
      }

      if (result.firstInvalidNode != null) {
        result.firstInvalidNode!.requestFocus();
      } else {
        SnackBarController.showIconAndText(
            Icons.report_gmailerrorred, "Eine Eingabe ist nicht gültig");
      }
      return false;
    }
  }

  Future<bool> _doSave(BuildContext context) async {
    var result = await onBeforeSave?.call(context) ?? true;

    var loading = LoadingWidgetService.instance.show();

    if (!result) {
      return false;
    }

    if (updateItemFrame.editableItem.id == null) {
      var itemRef = await databaseCollection.add(updateItemFrame.editableItem);
      onAfterInsert?.call(context, itemRef);
    } else {
      var subItemsChanged =
          (updateItemFrame.editableItem is HasDatabaseSubItems &&
                  updateItemFrame.originalItem is HasDatabaseSubItems)
              ? !(updateItemFrame.editableItem as HasDatabaseSubItems)
                  .equalSubItems(
                      updateItemFrame.originalItem as HasDatabaseSubItems?)
              : false;
      var itemRef = await databaseCollection.update(
        updateItemFrame.editableItem,
        supressSubItemUpdate: !subItemsChanged,
        notUpdateKeys: fieldsNotToUpdate,
      );

      if (itemRef == null) {
        loading.deactivate();
        await SplitRouterDelegate.instance.pushDialog(DefaultDialog(
          title: Text(
              "Die Verbindung ist fehlgeschlagen. Die Änderungen konnten nicht gespeichert werden."),
          buttons: [
            DefaultDialogButton(
              titel: Text("Verstanden"),
            )
          ],
        ));
        return false;
      }
      lockService?.removeLock(
          databaseCollection.collectionName, updateItemFrame.editableItem.id);

      onAfterSave?.call(context);
    }

    LockNavigatorService.instance.unlockTopStack();

    if (updateItemFrame.editableItem.id == null) {
      Navigator.of(context).pop();
    } else {
      updateItemFrame.setEditable(false);
      updateNotifier.update();
    }

    loading.deactivate();

    return true;
  }

  Future<bool> tryStartEditing(BuildContext context) async {
    FocusNode().requestFocus();
    var loading = LoadingWidgetService.instance.show();

    if (lockService != null) {
      try {
        var isLocked = await lockService!.isLocked(
          databaseCollection.collectionName,
          updateItemFrame.editableItem.id,
          catchExceptions: false,
        );
        if (isLocked) {
          loading.deactivate();
          await SplitRouterDelegate.instance.pushDialog(DefaultDialog(
            title: Text("Dieser Eintrag wird zurzeit bearbeitet"),
            buttons: [
              DefaultDialogButton(
                titel: Text("Verstanden"),
              )
            ],
          ));
          return false;
        }
      } catch (e) {
        loading.deactivate();
        SnackBarController.showText("Verbindung fehlgeschlagen");
        return false;
      }
    }

    lockService?.setLock(
        databaseCollection.collectionName, updateItemFrame.editableItem.id);

    LockNavigatorService.instance.lockForTopPage();
    updateItemFrame.setEditable(true);
    loading.deactivate();
    updateNotifier.update();
    return true;
  }

  Future<bool> tryDelete(BuildContext context) async {
    var loading = LoadingWidgetService.instance.show();

    if (lockService != null) {
      try {
        var isLocked = await lockService!.isLocked(
          databaseCollection.collectionName,
          updateItemFrame.editableItem.id,
          catchExceptions: false,
        );
        if (isLocked) {
          loading.deactivate();
          await SplitRouterDelegate.instance.pushDialog(DefaultDialog(
            title: Text(
                "Der Eintrag kann zurzeit nicht gelöscht werden, da er von einem anderen Benutzer bearbeitet wird."),
            buttons: [
              DefaultDialogButton(
                titel: Text("Verstanden"),
              )
            ],
          ));
          return false;
        }
      } catch (e) {
        loading.deactivate();
        SnackBarController.showText("Verbindung fehlgeschlagen");
        return false;
      }
    }

    loading.deactivate();
    var result = await onBeforeDelete?.call(context) ?? true;
    if (result == false) {
      return false;
    }

    bool confirmation = false;
    // await SplitRouterDelegate.instance.pushDialog(DefaultDialog(
    //   title: Text("Soll der Eintrag wirklich entgültig gelöscht werden?"),
    //   buttons: [
    //     DefaultDialogButton(titel: Text("Abbrechen"), onTap: () {}),
    //     DefaultDialogButton(
    //       titel: Text("Löschen"),
    //       onTap: () => confirmation = true,
    //     ),
    //   ],
    // ));

    await SplitRouterDelegate.instance.pushDialog(DeleteQuestionDialog(
      onSuccess: () => confirmation = true,
    ));
    if (!confirmation) return false;

    Navigator.of(context).pop();

    loading = LoadingWidgetService.instance.show();
    result = await databaseCollection.delete(updateItemFrame.editableItem);
    loading.deactivate();

    if (result == false) {
      SnackBarController.showText("Verbindung fehlgeschlagen");
    } else {
      SnackBarController.showIconAndText(
          Icons.delete_outline, "Erfolgreich gelöscht");
    }
    return true;
  }

  Future<void> abortEditing(BuildContext context) async {
    FocusNode().requestFocus();

    onAbort?.call(context);

    if (lockService != null) {
      lockService!.removeLock(
          databaseCollection.collectionName, updateItemFrame.editableItem.id);
    }
    LockNavigatorService.instance.unlockTopStack();

    if (updateItemFrame.editableItem.id == null) {
      Navigator.of(context).pop();
    } else {
      updateItemFrame.setEditable(false);
      updateNotifier.update();
    }
  }
}

abstract class ILockService {
  Future<bool> isLocked(
    String collectionName,
    String? itemId, {
    bool catchExceptions = true,
  });
  void setLock(
    String collectionName,
    String? itemId, {
    bool catchExceptions = true,
  });
  void removeLock(
    String collectionName,
    String? itemId, {
    bool catchExceptions = true,
  });
}

==> clean_android_design-0.2.18.tar.gz: lib/src/structure/update/update_controller.dart <==
import 'package:clean_android_design/clean_android_design.dart';
import 'package:clean_android_design/src/dialog/delete_question_dialog.dart';
import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:flutter/material.dart';

class UpdateController<T extends UpdateableDatabaseItem<T>>
    extends InheritedWidget {
  final UpdateItemFrame<T> updateItemFrame;
  final Validator<T> _validator;
  final ForceUpdateKey updateNotifier;

  final IUpdateScaffoldWidgetService<T> databaseCollection;
  final ILockService? lockService;

  final Future<bool> Function(BuildContext context)? onBeforeSave;
  final void Function(BuildContext context)? onAfterSave;
  final void Function(BuildContext context, DocumentReference? itemReference)?
      onAfterInsert;
  final void Function(BuildContext context)? onAbort;
  final Future<bool> Function(BuildContext context)? onBeforeDelete;
  final List<String>? fieldsNotToUpdate;

  ReadOnlyValidator<T> get validator => _validator.readOnly();

  UpdateController({
    required Widget child,
    required this.updateItemFrame,
    required Validator<T> validator,
    required this.databaseCollection,
    required this.onBeforeSave,
    required this.onAfterSave,
    required this.onAfterInsert,
    required this.onAbort,
    required this.fieldsNotToUpdate,
    required this.onBeforeDelete,
    required this.lockService,
  })  : _validator = validator,
        updateNotifier = ForceUpdateKey(),
        super(child: child);

  @override
  bool updateShouldNotify(InheritedWidget oldWidget) => true;

  static UpdateController<T> of<T extends UpdateableDatabaseItem<T>>(
      BuildContext context) {
    final UpdateController<T>? result =
        context.dependOnInheritedWidgetOfExactType<UpdateController<T>>();
    assert(result != null, 'No UpdateController found in context');
    return result!;
  }

  static UpdateController<T>? maybeOf<T extends UpdateableDatabaseItem<T>>(
      BuildContext context) {
    return context.dependOnInheritedWidgetOfExactType<UpdateController<T>>();
  }

  Future<bool> trySave(BuildContext context) async {
    FocusNode().requestFocus();

    var result = _validator.isValid(updateItemFrame.editableItem);
    if (result.isValid) {
      return _doSave(context);
    } else {
      if (result.firstInvalidPage != null &&
          ScaffoldPageController.isIn(context)) {
        ScaffoldPageController.of(context).setPage(result.firstInvalidPage!);
        await Future.delayed(Duration(milliseconds: 100));
        result = _validator.isValid(updateItemFrame.editableItem);
      }

      if (result.firstInvalidNode != null) {
        result.firstInvalidNode!.requestFocus();
      } else {
        SnackBarController.showIconAndText(
            Icons.report_gmailerrorred, "Eine Eingabe ist nicht gültig");
      }
      return false;
    }
  }

  Future<bool> _doSave(BuildContext context) async {
    var result = await onBeforeSave?.call(context) ?? true;

    var loading = LoadingWidgetService.instance.show();

    if (!result) {
      return false;
    }

    if (updateItemFrame.editableItem.id == null) {
      var itemRef = await databaseCollection.add(updateItemFrame.editableItem);
      onAfterInsert?.call(context, itemRef);
    } else {
      var subItemsChanged =
          (updateItemFrame.editableItem is HasDatabaseSubItems &&
                  updateItemFrame.originalItem is HasDatabaseSubItems)
              ? !(updateItemFrame.editableItem as HasDatabaseSubItems)
                  .equalSubItems(
                      updateItemFrame.originalItem as HasDatabaseSubItems?)
              : false;
      var itemRef = await databaseCollection.update(
        updateItemFrame.editableItem,
        supressSubItemUpdate: !subItemsChanged,
        notUpdateKeys: fieldsNotToUpdate,
      );

      if (itemRef == null) {
        loading.deactivate();
        await SplitRouterDelegate.instance.pushDialog(DefaultDialog(
          title: Text(
              "Die Verbindung ist fehlgeschlagen. Die Änderungen konnten nicht gespeichert werden."),
          buttons: [
            DefaultDialogButton(
              titel: Text("Verstanden"),
            )
          ],
        ));
        return false;
      }
      lockService?.removeLock(
          databaseCollection.collectionName, updateItemFrame.editableItem.id);

      onAfterSave?.call(context);
    }

    LockNavigatorService.instance.unlockTopStack();

    if (updateItemFrame.editableItem.id == null) {
      Navigator.of(context).pop();
    } else {
      updateItemFrame.setEditable(false);
      updateNotifier.update();
    }

    loading.deactivate();

    return true;
  }

  Future<bool> tryStartEditing(BuildContext context) async {
    FocusNode().requestFocus();
    var loading = LoadingWidgetService.instance.show();

    if (lockService != null) {
      try {
        var isLocked = await lockService!.isLocked(
          databaseCollection.collectionName,
          updateItemFrame.editableItem.id,
          catchExceptions: false,
        );
        if (isLocked) {
          loading.deactivate();
          await SplitRouterDelegate.instance.pushDialog(DefaultDialog(
            title: Text("Dieser Eintrag wird zurzeit bearbeitet"),
            buttons: [
              DefaultDialogButton(
                titel: Text("Verstanden"),
              )
            ],
          ));
          return false;
        }
      } catch (e) {
        loading.deactivate();
        SnackBarController.showText("Verbindung fehlgeschlagen");
        return false;
      }
    }

    lockService?.setLock(
        databaseCollection.collectionName, updateItemFrame.editableItem.id);

    LockNavigatorService.instance.lockForTopPage();
    updateItemFrame.setEditable(true);
    loading.deactivate();
    updateNotifier.update();
    return true;
  }

  Future<bool> tryDelete(BuildContext context,
      {List<String>? whatsDeleted}) async {
    var loading = LoadingWidgetService.instance.show();

    if (lockService != null) {
      try {
        var isLocked = await lockService!.isLocked(
          databaseCollection.collectionName,
          updateItemFrame.editableItem.id,
          catchExceptions: false,
        );
        if (isLocked) {
          loading.deactivate();
          await SplitRouterDelegate.instance.pushDialog(DefaultDialog(
            title: Text(
                "Der Eintrag kann zurzeit nicht gelöscht werden, da er von einem anderen Benutzer bearbeitet wird."),
            buttons: [
              DefaultDialogButton(
                titel: Text("Verstanden"),
              )
            ],
          ));
          return false;
        }
      } catch (e) {
        loading.deactivate();
        SnackBarController.showText("Verbindung fehlgeschlagen");
        return false;
      }
    }

    loading.deactivate();
    var result = await onBeforeDelete?.call(context) ?? true;
    if (result == false) {
      return false;
    }

    bool confirmation = false;
    // await SplitRouterDelegate.instance.pushDialog(DefaultDialog(
    //   title: Text("Soll der Eintrag wirklich entgültig gelöscht werden?"),
    //   buttons: [
    //     DefaultDialogButton(titel: Text("Abbrechen"), onTap: () {}),
    //     DefaultDialogButton(
    //       titel: Text("Löschen"),
    //       onTap: () => confirmation = true,
    //     ),
    //   ],
    // ));

    await SplitRouterDelegate.instance.pushDialog(DeleteQuestionDialog(
      whatsDeleted: whatsDeleted,
      onSuccess: () => confirmation = true,
    ));
    if (!confirmation) return false;

    Navigator.of(context).pop();

    loading = LoadingWidgetService.instance.show();
    result = await databaseCollection.delete(updateItemFrame.editableItem);
    loading.deactivate();

    if (result == false) {
      SnackBarController.showText("Verbindung fehlgeschlagen");
    } else {
      SnackBarController.showIconAndText(
          Icons.delete_outline, "Erfolgreich gelöscht");
    }
    return true;
  }

  Future<void> abortEditing(BuildContext context) async {
    FocusNode().requestFocus();

    onAbort?.call(context);

    if (lockService != null) {
      lockService!.removeLock(
          databaseCollection.collectionName, updateItemFrame.editableItem.id);
    }
    LockNavigatorService.instance.unlockTopStack();

    if (updateItemFrame.editableItem.id == null) {
      Navigator.of(context).pop();
    } else {
      updateItemFrame.setEditable(false);
      updateNotifier.update();
    }
  }
}

abstract class ILockService {
  Future<bool> isLocked(
    String collectionName,
    String? itemId, {
    bool catchExceptions = true,
  });
  void setLock(
    String collectionName,
    String? itemId, {
    bool catchExceptions = true,
  });
  void removeLock(
    String collectionName,
    String? itemId, {
    bool catchExceptions = true,
  });
}
